/*{ cm3_mutex.c ***********************************************
 *
 * Written by Gregor Rebel 2010-2012
 *
 * Architecture independent support for inter task communication & synchronization.
 * 
 * Currently supported schedulers: FreeRTOS
}*/
#include "cm3_mutex.h"

//{ functions *************************************************

       cm3_Mutex* cm3_mutex_create() {

    cm3_Mutex* NewMutex = (cm3_Mutex*) ttc_memory_alloc( 1 );
    Assert_MUTEXES(NewMutex, ec_Malloc);

    return NewMutex;
}
             void cm3_mutex_init(cm3_Mutex* Lock) {
    // pretty simple
    *Lock = 0;
}
ttc_mutex_error_e cm3_mutex_lock(cm3_Mutex* Mutex, Base_t TimeOut) {
    Assert_MUTEXES(Mutex != 0, ec_InvalidArgument);

    if ( cm3_mutex_trylock(Mutex) == tme_OK ) // general case first: try to obtain lock
        return tme_OK;                        // got lock: simple and fast

    if      (TimeOut == -1) { // lock with endless timeout
        // wait until lock is free
        while (cm3_mutex_trylock(Mutex) != tme_OK)
            cm3_mutex_wait();

        return tme_OK;
    }
    else if (TimeOut ==  0) // no timeout: we already tried to obtain mutex without timeout
        return tme_MutexAlreadyLocked;
    else {                    // lock within timeout
        Base_t CurrentTime = ttc_task_get_tick_count();
        Base_t EndTime     = CurrentTime + TimeOut;
        ttc_mutex_error_e LockFailed = tme_MutexAlreadyLocked;

        while ( (EndTime < CurrentTime) && // tick-timer can overflow during wait: wait until overflow
                LockFailed                 // or until successfull lock
                ) {
            LockFailed = cm3_mutex_trylock(Mutex);
            if (LockFailed) {
                CurrentTime = ttc_task_get_tick_count();
                cm3_mutex_wait();
            }
        }

        while ( LockFailed &&
                (CurrentTime < EndTime)
              ) {         // loop until successfull lock or timeout
            LockFailed = cm3_mutex_trylock(Mutex);
            if (LockFailed) {
                CurrentTime = ttc_task_get_tick_count();
                cm3_mutex_wait();
            }
        };

        if (!LockFailed)
            return tme_OK;

        return tme_TimeOut;
    }

    return tme_UNKNOWN; // this line should not be reachable
}
             void cm3_mutex_wait() {
    ttc_task_yield();
}
ttc_mutex_error_e cm3_mutex_trylock(cm3_Mutex* Mutex) {
    Assert_MUTEXES(Mutex != 0, ec_InvalidArgument);

    if ( __LDREXB(Mutex) )
        return tme_MutexAlreadyLocked;

    if ( __STREXB(1, Mutex)) {
        __CLREX(); // remove lock created by __LDREXB()
        return tme_MutexAlreadyLocked;
    }

    __CLREX(); // remove lock created by __LDREXB()
    return tme_OK;
}
ttc_mutex_error_e cm3_mutex_unlock(cm3_Mutex* Mutex) {
    Assert_MUTEXES(Mutex != 0, ec_InvalidArgument);
    Assert_MUTEXES(*Mutex, ec_InvalidImplementation); // unlocking a mutex that is not locked!

    *Mutex = 0;

    return tme_NotImplemented;
}

// private functions ********************************************

//}functions
