#ifndef TTC_GPIO_H
#define TTC_GPIO_H

/*{ ttc_gpio.h ***********************************************
 
 Written by Gregor Rebel 2010-2012

 Architecture independent support for General Purpose Input Output (GPIO).

 Currently implemented architectures: stm32


 Example usage:
 
 Makefile
    COMPILE_OPTS += -DTTC_LED1=PIN_PA8
 
 Source
    #include "ttc_gpio.h"
 
    // use of constant gpios:
    ttc_gpio_init(TTC_LED1, tgm_output_push_pull);
    while (1) {
      ttc_gpio_set(TTC_LED1);
      ttc_gpio_clr(TTC_LED1);
    }

    // use of variable based gpios:
    ttc_Port_t LED;
    ttc_gpio_init(&LED, TTC_LED1, tgm_output_push_pull);
    while (1) {
      ttc_gpio_set_variable(&Port);
      ttc_gpio_clr_variable(&Port);
    }

    // use of shortcuts:
    while (1) {
      gset(TTC_LED1);
      gclr(TTC_LED1);
    }
    while (1) {
      gset_v(&Port);
      gclr_v(&Port);
    }

    // 8-bit wide parallel ports
    #ifdef TTC_PARALLEL08_1
      for (int Bit = 0; Bit < 8; Bit++) {
          ttc_gpio_init2(TTC_PARALLEL08_1_BANK, Bit + TTC_PARALLEL08_1_FIRSTPIN,  tgm_output_push_pull, tgs_Max);
      }
      u8_t Count = 0;
      while (1) {
        ttc_gpio_set_u8_pin3(TTC_PARALLEL08_1_BANK, TTC_PARALLEL08_1_FIRSTPIN, Count++);
      }
    #endif

    // 16-bit wide parallel ports
    #ifdef TTC_PARALLEL16_1
      for (int Bit = 0; Bit < 16; Bit++) {
          ttc_gpio_init2(TTC_PARALLEL16_1_BANK, Bit,  tgm_output_push_pull, tgs_Max);
      }
      u8_t Count = 0;
      while (1) {
        ttc_gpio_set_u16(TTC_PARALLEL16_1_BANK, Count++);
      }
    #endif
}*/
//{ includes

#include "ttc_basic.h"
#include "ttc_gpio_types.h"

//}includes
//{ Prototypes

/** initializes single port bit for use as input/ output with given speed by use of a state variable
 *
 * @param Port   points to structure to be filled with given data
 * @param Bank  GPIO bank to use (-> ttc_gpio_types.h:ttc_GPIO_BANK)
 *               Note: Bank, Pin can be given as one argument by use of Pin_Pxn macro (GPIOA, 7) = Pin_PA7
 * @param Pin    0..15
 * @param Type   mode to use (-> ttc_gpio_types.h:ttc_gpio_mode_e)
 * @param Speed  speed to use (-> ttc_gpio_types.h:ttc_gpio_speed_e)
 */
void ttc_gpio_init2_variable(ttc_Port_t* Port, ttc_GPIO_BANK Bank, u8_t Pin, ttc_gpio_mode_e Type, ttc_gpio_speed_e Speed);

#define ttc_gpio_init_variable(Port, Bank_Pin, Type) ttc_gpio_init2_variable(Port, Bank_Pin, Type, tgs_Max);

//} Prototypes
/*{ Function Redefines

  Most functions in ttc_gpio are just redefinitions of functions or macros implemented in low-level drivers.
  These macro definitions can provide a faster access to individual port bits.
  If you want to realize parallel busses, take a look at the u8 and u16 access functions.

  On most architectures, gpio operations require only some few assembly instructions.
  Adding another function call with argument passing can reduce maximum
  switching speed by a factor >4.

  You can see descriptions of equivalens functions below.
*/

/** initializes single port bit for use as input/ output with maximum speed by use of a state variable
 *
 * @param Port   points to structure to be filled with given data
 * @param Bank  GPIO bank to use (-> ttc_gpio_types.h:ttc_GPIO_BANK)
 *               Note: Bank, Pin can be given as one argument by use of Pin_Pxn macro (GPIOA, 7) = Pin_PA7
 * @param Pin    0..15
 * @param Type   mode to use (-> ttc_gpio_types.h:ttc_gpio_mode_e)
 */
// void ttc_gpio_init_variable(ttc_Port_t* Port, ttc_GPIO_BANK Bank, u8_t Pin, ttc_gpio_mode_e Type);

/** allows to load gpio port data into a variable for use in later port operations
 *
 * @param Port   points to structure to be initialized by ttc_gpio_init2_variable()
 * @param Bank  one of GPIOA, GPIOB, GPIOC, GPIOD, GPIOE, GPIOF
 *               Note: Bank, Pin can be given as one argument by use of Pin_Pxn macro (GPIOA, 7) = Pin_PA7
 * @param Pin    0..15
 * @return  filled out struct *Port
 *
 * Example:
 * Makefile
 *   COMPILE_OPTS += -DTTC_LED1=PIN_PA7
 *
 * Source
 *   ttc_Port_t Port;
 *   ttc_gpio_variable(&Port, TTC_LED1);
 *   ttc_gpio_init_variable(Port.Bank, Port.Pin, tgm_output_push_pull);
 */
// void ttc_gpio_variable(ttc_Port_t* Port, ttc_GPIO_BANK Bank, u8_t Pin);

/** initializes single port bit for use as input/ output with maximum speed
 *
 * @param Bank  GPIO bank to use (-> ttc_gpio_types.h:ttc_GPIO_BANK)
 *               Note: Bank, Pin can be given as one argument by use of Pin_Pxn macro (GPIOA, 7) = Pin_PA7
 * @param Pin    0..15
 * @param Type   mode to use (-> ttc_gpio_types.h:ttc_gpio_mode_e)
 */
void ttc_gpio_init(ttc_GPIO_BANK Bank, u8_t Pin, ttc_gpio_mode_e Type);

/** initializes single port bit for use as input/ output with given speed
 *
 * @param Bank  GPIO bank to use (-> ttc_gpio_types.h:ttc_GPIO_BANK)
 *               Note: Bank, Pin can be given as one argument by use of Pin_Pxn macro (GPIOA, 7) = Pin_PA7
 * @param Pin    0..15
 * @param Type   mode to use (-> ttc_gpio_types.h:ttc_gpio_mode_e)
 */
void ttc_gpio_init2(ttc_GPIO_BANK Bank, u8_t Pin, ttc_gpio_mode_e Type, ttc_gpio_speed_e Speed);

/** read current value of input pin by use of preinitialized port variable
 *
 * @param Port  filled out struct as being returned by ttc_gpio_variable()
 * @return      =1: voltage at input pin is logical 1; logical 0 otherwise
 */
// bool ttc_gpio_get_variable(ttc_Port_t* Port);

/** set output pin to logical one by use of preinitialized port variable
 *
 * @param Port  filled out struct as being returned by ttc_gpio_variable()
 */
// void ttc_gpio_set_variable(ttc_Port_t* Port);

/** set output pin to logical zero by use of preinitialized port variable
 *
 * @param Port  filled out struct as being returned by ttc_gpio_variable()
 */
// void ttc_gpio_clr_variable(ttc_Port_t* Port);

/** sets output port to given value
 *
 * @param Port  filled out struct as being returned by ttc_gpio_variable()
 * @param Value  =0: set output to logical 0; to logical 1 otherwise
 */
// void ttc_gpio_setn_variable(ttc_Port_t* Port, BOOL Value);

/** creates numeric 8-bit representing given GPIO-bank and pin number
 *
 * @param Bank  GPIO bank to use (-> ttc_gpio_types.h:ttc_GPIO_BANK)
 *               Note: Bank, Pin can be given as one argument by use of Pin_Pxn macro (GPIOA, 7) = Pin_PA7
 * @param Pin    0..15
 * @return       8-bit value representing given GPIO-pin in a memory friendly format
 */
// u8_t ttc_gpio_create_index8(ttc_GPIO_BANK Bank, u8_t Pin);

/** restores GPIO-bank and pin number from 8-bit value
 *
 * @param PhysicalIndex  8-bit value representing GPIO-bank + pin
 * @param Bank          loaded with GPIO bank (-> ttc_gpio_types.h:ttc_GPIO_BANK)
 * @param Pin            loaded with pin number
 */
// void ttc_gpio_from_index8(u8_t PhysicalIndex, ttc_GPIO_BANK* Bank, u8_t* Pin);

/** reads current value of single port bit (if initialized as input)
 *
 * Note: This macro combines Bank+Pin into one argument.
 *       If the compiler complains about the amount of arguments, one can use  ttc_gpio_get2() instead.
 *
 * @param Bank+Pin  Bank and Pin given as one argument by use of Pin_Pxn macro: E.g. (GPIOA, 7) = Pin_PA7
 * @param Bank      GPIO bank to use (-> ttc_gpio_types.h:ttc_GPIO_BANK)
 * @param Pin       0..15
 * @param Type      mode to use (-> ttc_gpio_types.h:ttc_gpio_mode_e)
 * @return       =1: voltage at input pin is logical 1; logical 0 otherwise
 */
//bool ttc_gpio_get(Bank+Pin);
//bool ttc_gpio_get2(ttc_GPIO_BANK Bank, u8_t Pin);

/** sets output port to logical 1
 *
 * Note: This macro combines Bank+Pin into one argument.
 *       If the compiler complains about the amount of arguments, one can use  ttc_gpio_set2() instead.
 *
 * @param Bank+Pin  Bank and Pin given as one argument by use of Pin_Pxn macro: E.g. (GPIOA, 7) = Pin_PA7
 * @param Bank      GPIO bank to use (-> ttc_gpio_types.h:ttc_GPIO_BANK)
 * @param Pin       0..15
 * @param Pin    0..15
 */
//void ttc_gpio_set(Bank+Pin);
//void ttc_gpio_set2(ttc_GPIO_BANK Bank, u8_t Pin);

/** sets output port to logical 0
 *
 * Note: This macro combines Bank+Pin into one argument.
 *       If the compiler complains about the amount of arguments, one can use  ttc_gpio_clr2() instead.
 *
 * @param Bank+Pin  Bank and Pin given as one argument by use of Pin_Pxn macro: E.g. (GPIOA, 7) = Pin_PA7
 * @param Bank      GPIO bank to use (-> ttc_gpio_types.h:ttc_GPIO_BANK)
 * @param Pin       0..15
 */
//void ttc_gpio_clr(Bank+Pin);
//void ttc_gpio_clr2(ttc_GPIO_BANK Bank, u8_t Pin);

/** sets output port to given value
 *
 * Note: This macro combines Bank+Pin into one argument.
 *       If the compiler complains about the amount of arguments, one can use  ttc_gpio_setn3() instead.
 *
 * @param Bank+Pin  Bank and Pin given as one argument by use of Pin_Pxn macro: E.g. (GPIOA, 7) = Pin_PA7
 * @param Bank      GPIO bank to use (-> ttc_gpio_types.h:ttc_GPIO_BANK)
 * @param Pin       0..15
 * @param Type      mode to use (-> ttc_gpio_types.h:ttc_gpio_mode_e)
 * @param Value     =0: set output to logical 0; to logical 1 otherwise
 */
//void ttc_gpio_setn(Bank+Pin, BOOL Value);
//void ttc_gpio_setn3(ttc_GPIO_BANK Bank, u8_t Pin, BOOL Value);

/** initialize 16 bit wide parallel port
 *
 * @param Bank+FirstPin  Bank and Pin given as one argument by use of Pin_Pxn macro: E.g. (GPIOA, 7) = Pin_PA7
 * @param Bank           GPIO bank to use (GPIO_BANKA, GPIO_BANKB, ...)
 * @param Type           mode to use for all 16 pins  (-> ttc_gpio_types.h:ttc_gpio_mode_e)
 * @param Speed          speed to use for all 16 pins (-> ttc_gpio_types.h:ttc_gpio_speed_e)
 */
// void ttc_gpio_init_u16(Bank, Type, Speed);
void _ttc_gpio_init_u16(ttc_GPIO_BANK Bank, ttc_gpio_mode_e Type, ttc_gpio_speed_e Speed);

/** read 16 bit value from gpio bank at once (fast access)
 *
 * @param Bank     GPIO bank to use (GPIO_BANKA, GPIO_BANKB, ...)
 */
// u16_t ttc_gpio_get_u16(ttc_GPIO_BANK Bank);
u16_t _ttc_gpio_get_u16(ttc_GPIO_BANK Bank);

/** write 16 bit value into gpio bank at once (fast access)
 *
 * @param Bank     GPIO bank to use (GPIO_BANKA, GPIO_BANKB, ...)
 */
// void ttc_gpio_set_u16(ttc_GPIO_BANK Bank, u16_t Value);
void _ttc_gpio_set_u16(ttc_GPIO_BANK Bank, u16_t Value);

/** initialize 8 bit wide parallel port
 *
 * @param Bank+FirstPin  Bank and Pin given as one argument by use of Pin_Pxn macro: E.g. (GPIOA, 7) = Pin_PA7
 * @param Bank           GPIO bank to use (GPIO_BANKA, GPIO_BANKB, ...)
 * @param FirstPin       0..8  lowest of 8 consecutive pins to use
 * @param Type           mode to use for all 8 pins  (-> ttc_gpio_types.h:ttc_gpio_mode_e)
 * @param Speed          speed to use for all 8 pins (-> ttc_gpio_types.h:ttc_gpio_speed_e)
 */
// defined as macros:
// void ttc_gpio_init_u8_pin(Bank_FirstPin, Type, Speed);
// void ttc_gpio_init_u8_pin4(Bank, FirstPin, Type, Speed);
void _ttc_gpio_init_u8_pin(ttc_GPIO_BANK Bank, u8_t FirstPin, ttc_gpio_mode_e Type, ttc_gpio_speed_e Speed);

/** read 8 bit value starting at certain pin from gpio bank
 *
 * @param Bank+FirstPin  Bank and Pin given as one argument by use of Pin_Pxn macro: E.g. (GPIOA, 7) = Pin_PA7
 * @param Bank           GPIO bank to use (GPIO_BANKA, GPIO_BANKB, ...)
 * @param Pin            0..8  lowest of 8 consecutive pins to use
 * @return               bits Pin..Pin+7 from input data register of given gpio bank
 */
// u8_t ttc_gpio_get_u8_pin(Bank+FirstPin);
// u8_t ttc_gpio_get_u8_pin2(ttc_GPIO_BANK Bank, u8_t Pin);
u8_t _ttc_gpio_get_u8_pin(ttc_GPIO_BANK Bank, u8_t Pin);

/** write 8 bit value starting at certain pin into gpio bank
 *
 * @param Bank+FirstPin  Bank and Pin given as one argument by use of Pin_Pxn macro: E.g. (GPIOA, 7) = Pin_PA7
 * @param Bank           GPIO bank to use (GPIO_BANKA, GPIO_BANKB, ...)
 * @param FirstPin       0..8  lowest of 8 consecutive pins to use
 * @param Value          written into bits Pin..Pin+7 of output data register of given gpio bank
 */
// void ttc_gpio_set_u8_pin(Bank+FirstPin, u8_t Value);
// void ttc_gpio_set_u8_pin2(ttc_GPIO_BANK Bank, u8_t FirstPin, u8_t Value);
void _ttc_gpio_set_u8_pin(ttc_GPIO_BANK Bank, u8_t FirstPin, u8_t Value);

/* Macro Arguments
 *
 * PORT      pointer to type ttc_Port_t
 * BANK      pointer to type ttc_GPIO_BANK
 * PIN       0..15 pin number in GPIO bank
 * FIRST_PIN 0..8 pin number of first pin of am 8-bit wide parallel port
 * TYPE      type to configure pin to  (one from ttc_gpio_types.h:ttc_gpio_mode_e)
 * SPEED     speed to configure pin to (one from ttc_gpio_types.h:ttc_gpio_speed_e)
 * BANK_PIN  BANK,PIN  (allows use of PIN_Pxn definitions like PIN_A1)
 * INDEX     u8_t Value storing GPIO-Bank index (0..15) and Pin index (0..15) in high- and low-nibble
 *
 */

#define ttc_gpio_setn(BANK_PIN,VALUE)      if (VALUE) ttc_gpio_set2(BANK_PIN);     else ttc_gpio_clr2(BANK_PIN)
#define ttc_gpio_setn3(BANK,PIN,VALUE)     if (VALUE) ttc_gpio_set2(BANK,PIN);     else ttc_gpio_clr2(BANK,PIN)
#define ttc_gpio_setn_variable(PORT,VALUE) if (VALUE) ttc_gpio_set_variable(PORT); else ttc_gpio_clr_variable(PORT)

//} Function Redefines
//{ check for missing low-level implementations

#ifndef ttc_gpio_get
  #warning Missing low level implementation for ttc_gpio_get()
#endif
#ifndef ttc_gpio_set
  #warning Missing low level implementation for ttc_gpio_set()
#endif
#ifndef ttc_gpio_clr
  #warning Missing low level implementation for ttc_gpio_clr()
#endif
#ifndef ttc_gpio_create_index8
  #warning Missing low level implementation for ttc_gpio_create_index8()
#endif
#ifndef ttc_gpio_from_index8
  #warning Missing low level implementation for ttc_gpio_from_index8()
#endif
#ifndef ttc_gpio_get_variable
  #warning Missing low level implementation for ttc_gpio_get_variable()
#endif
#ifndef ttc_gpio_set_variable
  #warning Missing low level implementation for ttc_gpio_set_variable()
#endif
#ifndef ttc_gpio_clr_variable
  #warning Missing low level implementation for ttc_gpio_clr_variable()
#endif
#ifndef ttc_gpio_variable
  #warning Missing low level implementation for ttc_gpio_variable()
#endif

//}
//{ replace some missing low-level implementations by slower high-level implementations

#ifndef ttc_gpio_get_u8_pin
#define ttc_gpio_get_u8_pin(FIRST_PORT)                    _ttc_gpio_get_u8_pin(FIRST_PORT)
#define ttc_gpio_get_u8_pin2(BANK,FIRST_PIN)               _ttc_gpio_get_u8_pin(BANK,FIRST_PIN)
#endif
#ifndef ttc_gpio_set_u8_pin
#define ttc_gpio_set_u8_pin(FIRST_PORT,VALUE)              _ttc_gpio_set_u8_pin(FIRST_PORT,VALUE)
#define ttc_gpio_set_u8_pin3(BANK,FIRST_PIN,VALUE)         _ttc_gpio_set_u8_pin(BANK,FIRST_PIN,VALUE)
#endif
#ifndef ttc_gpio_get_u16
#define ttc_gpio_get_u16(BANK)                             _ttc_gpio_get_u16(BANK)
#endif
#ifndef ttc_gpio_set_u16
#define ttc_gpio_set_u16(BANK,VALUE)                       _ttc_gpio_set_u16(BANK,VALUE)
#endif
#ifndef ttc_gpio_init_u8_pin
#define ttc_gpio_init_u8_pin(BANK_PIN,TYPE,SPEED)          _ttc_gpio_init_u8_pin(BANK_PIN,TYPE,SPEED)
#endif
#ifndef ttc_gpio_init_u8_pin4
#define ttc_gpio_init_u8_pin4(BANK,FIRST_PIN,TYPE,SPEED)   _ttc_gpio_init_u8_pin(BANK,FIRST_PIN,TYPE,SPEED)
#endif
#ifndef ttc_gpio_init_u16
#define ttc_gpio_init_u16(BANK,TYPE,SPEED)                 _ttc_gpio_init_u16(BANK,TYPE,SPEED)
#endif

//}
//{ shortcuts

#define gget   ttc_gpio_get
#define gset   ttc_gpio_set
#define gclr   ttc_gpio_clr
#define gget_v ttc_gpio_get_variable
#define gset_v ttc_gpio_set_variable
#define gclr_v ttc_gpio_clr_variable

//}
#endif
