#ifndef TTC_I2C_H
#define TTC_I2C_H

/*{ ttc_i2c.h **********************************************************
 
                      The ToolChain
                      
   Device independent support for 
   Serial Peripheral Interface (I2C)
   
   Currently implemented architectures: stm32

   
   written by Gregor Rebel 2012
   
   Reference:
   
   RM008 - STM32 Reference Manual p. 672
   -> http://www.st.com/internet/com/TECHNICAL_RESOURCES/TECHNICAL_LITERATURE/REFERENCE_MANUAL/CD00171190.pdf
   
   
   How it works
   
   Each makefile.100_board* can define one or more I2Cs being routed to connectors.
   This is done via lines of this type:
     COMPILE_OPTS += -D<OPTION>=<VALUE>
   This sets compile option <OPTION> to value <VALUE>
   
   I2Cn are enumerated starting at index 1 (e.g. I2C1, I2C2, ..).
   Each I2Cn can be configured via these constants:
   TTC_I2C1       <I2C>                  sets internal I2C device to use
   TTC_I2C1_MOSI  <GPIO_BANK>,<PIN_NO>   sets port to use for Master Out Slave In pin
   TTC_I2C1_MISO  <GPIO_BANK>,<PIN_NO>   sets port to use for Master In Slave Out pin
   TTC_I2C1_SCK   <GPIO_BANK>,<PIN_NO>   sets port to use for Serial Clock pin
   TTC_I2C1_NSS   <GPIO_BANK>,<PIN_NO>   sets port to use for Slave Select pin
   
   # Example excerpt from makefile.100_board_olimex_p107
   COMPILE_OPTS += -DTTC_I2C1=I2C3          # I2C connected to RS232 10-pin male header UEXT and SDCARD slot
   COMPILE_OPTS += -DTTC_I2C1_MOSI=GPIOC,12 # UEXT pin 8
   COMPILE_OPTS += -DTTC_I2C1_MISO=GPIOC,11 # UEXT pin 7
   COMPILE_OPTS += -DTTC_I2C1_SCK=GPIOC,10  # UEXT pin 9
   #COMPILE_OPTS += -DTTC_I2C1_NSS=         # NSS not connected 
    

}*/
//{ Defines/ TypeDefs ****************************************************

//} Defines
//{ Includes *************************************************************

#include "ttc_basic.h"
#include "ttc_i2c_types.h"

#ifdef TARGET_ARCHITECTURE_STM32
  #include "stm32_i2c.h"
#endif

//} Includes
//{ Structures/ Enums 2 **************************************************

// architecture specific configuration data of single I2C 
#ifdef TARGET_ARCHITECTURE_STM32
    typedef stm32_i2c_architecture_t ttc_i2c_architecture_t;
#endif

//} Structures/ Enums
//{ Global Variables *****************************************************



//} Global Variables
//{ Function prototypes **************************************************


/* returns amount of I2C devices available on current uC
 * @return amount of available I2C devices
 */
u8_t ttc_i2c_get_max_index();

/* fills out given I2C_Generic with default values for indexed I2C
 * @param I2C_Index     device index of I2C to init (1..ttc_i2c_get_max_index())
 * @param I2C_Generic   pointer to struct ttc_i2c_generic_t
 * @return  == 0:         *I2C_Generic has been initialized successfully; != 0: error-code
 */
ttc_i2c_errorcode_e ttc_i2c_get_defaults(u8_t I2C_Index, ttc_i2c_generic_t* I2C_Generic);

/* fills out given I2C_Generic with maximum valid values for indexed I2C
 * @param I2C_Index     device index of I2C to init (1..ttc_i2c_get_max_index())
 * @param I2C_Generic   pointer to struct ttc_i2c_generic_t
 * @return  == 0:         *I2C_Generic has been initialized successfully; != 0: error-code
 */
ttc_i2c_errorcode_e ttc_i2c_get_features(u8_t I2C_Index, ttc_i2c_generic_t* I2C_Generic);

/* initializes single I2C
 * @param I2C_Index     device index of I2C to init (1..ttc_i2c_get_max_index() )
 * @param I2C_Generic   filled out struct ttc_i2c_generic_t                                    (Note: referenced struct must stay in memory as long as device is in use!)
 * @param I2C_Arch      empty structure which will be filled with architecture dependend data  (Note: referenced struct must stay in memory as long as device is in use!)
 * @return              == 0: I2C has been initialized successfully; != 0: error-code
 */
ttc_i2c_errorcode_e ttc_i2c_init(u8_t I2C_Index, ttc_i2c_generic_t* I2C_Generic, ttc_i2c_architecture_t* I2C_Arch);

/* Send out given byte to addressed target device
 *
 * Note: I2C must be initialized before!
 * Note: This function is partially thread safe (use it for each I2C from one thread only)!
 * Note: This function blocks until all bytes have been sent!
 *
 * @param I2C_Index      device index of I2C to init (1..ttc_i2c_get_max_index() )
 * @param TargetAddress  receivers own TargetAddress
 * @param Register       index of start register where to store bytes in receiving device
 * @param RegisterType   size and order of register adresses (amount of bytes to transfer for each register adress)
 * @param Byte           8-bit data to be send
 * @return               == 0: Buffer has been sent successfully; != 0: error-code
 */
ttc_i2c_errorcode_e ttc_i2c_write_register(u8_t I2C_Index, const u16_t TargetAddress, const u32_t Register, ttc_i2c_register_type_e RegisterType, const char Byte);

/* Send out given amount of bytes to addressed target device.
 *
 * Note: I2C must be initialized before!
 * Note: This function is partially thread safe (use it for each I2C from one thread only)!
 * Note: This function blocks until all bytes have been sent!
 * 
 * @param I2C_Index    device index of I2C to init (1..ttc_i2c_get_max_index() )
 * @param TargetAddress      receivers own TargetAddress
 * @param Register     index of start register where to store bytes in receiving device
 * @param RegisterType   size and order of register adresses (amount of bytes to transfer for each register adress)
 * @param Buffer       memory location from which to read data
 * @param Amount       amount of bytes to send from Buffer[]
 * @return             == 0: Buffer has been sent successfully; != 0: error-code
 */
ttc_i2c_errorcode_e ttc_i2c_write_registers(u8_t I2C_Index, const u16_t TargetAddress, const u32_t Register, ttc_i2c_register_type_e RegisterType, const u8_t* Buffer, u16_t Amount);


/* Reads single data byte from addressed target device.
 *
 * Note: I2C must be initialized before!
 * Note: This function is partially thread safe (use it for each I2C from one thread only)!
 * Note: This function blocks until a valid byte has been received or timeout occurs
 *
 * @param I2C_Index      device index of I2C to init (1..ttc_i2c_get_max_index() )
 * @param TargetAddress  receivers own TargetAddress
 * @param Register       index of start register where to store bytes in receiving device
 * @param RegisterType   size and order of register adresses (amount of bytes to transfer for each register adress)
 * @param Byte           pointer to 8 bit buffer where to store received byte
 * @param TimeOut        >0: max microseconds to wait; ==0: no timeout
 * @return               == 0: Byte has been read successfully; != 0: error-code
 */
ttc_i2c_errorcode_e ttc_i2c_read_register(u8_t I2C_Index, const u16_t TargetAddress, const u32_t Register, ttc_i2c_register_type_e RegisterType, u8_t* Buffer);

/* Reads amount of data bytes from addressed target device.
 *
 * Note: I2C must be initialized before!
 * Note: This function is partially thread safe (use it for each I2C from one thread only)!
 * Note: This function blocks until a valid byte has been received or timeout occurs
 *
 * @param I2C_Index      device index of I2C to init (1..ttc_i2c_get_max_index() )
 * @param TargetAddress  receivers own TargetAddress
 * @param Register       index of start register where to store bytes in receiving device
 * @param RegisterType   size and order of register adresses (amount of bytes to transfer for each register adress)
 * @param Byte           pointer to start of 8 bit buffer where to store received bytes
 * @param Amount         amount of bytes to be read from target device
 * @param AmountRead     will be loaded with amount of bytes being read
 * @param TimeOut        >0: max microseconds to wait; ==0: no timeout
 * @return               == 0: one or more bytes have been read successfully; != 0: error-code
 */
ttc_i2c_errorcode_e ttc_i2c_read_registers(u8_t I2C_Index, const u16_t TargetAddress, const u32_t Register, ttc_i2c_register_type_e RegisterType, u8_t* Buffer, u16_t Amount, u16_t* AmountRead);

//} Function prototypes

#endif //TTC_I2C_H
