/** ttc_mutex.c ***********************************************
 *
 * Written by Gregor Rebel 2010-2012
 *
 * Architecture independent support for inter task communication & synchronization.
 * 
 * Currently supported schedulers: FreeRTOS
 */
#include "ttc_mutex.h"

//{ global variables ********************************************

#if TTC_SMART_MUTEX_HISTORY_SIZE > 0

    // stores pointers to smart mutexes given for latest ttc_mutex_lock_smart() calls
    ttc_mutex_smart_t* tml_LastSmartMutex[TTC_SMART_MUTEX_HISTORY_SIZE];

    // history of smart mutexes without doubles
    ttc_mutex_smart_t* tml_UniqueSmartMutex[TTC_SMART_MUTEX_HISTORY_SIZE];

    // index of next valid entry inside tml_LastSmartMutex[]
    u8_t tml_Index_LastSmartMutex = 0;
    u8_t tml_Index_UniqueSmartMutex = 0;

    u32_t tml_AmountSmartLocks = 0;
#endif

//}global variables
//{ functions *************************************************

           ttc_Mutex* ttc_mutex_create() {
#ifdef DRIVER_ttc_mutex_create
    return DRIVER_ttc_mutex_create();
#else
    return NULL;
#endif
}
        ttc_mutex_smart_t* ttc_mutex_create_smart() {

    ttc_mutex_smart_t* NewMutex = ttc_memory_alloc( sizeof(ttc_mutex_smart_t) );
    NewMutex->Mutex = ttc_mutex_create();

    return NewMutex;
}
                      void ttc_mutex_init(ttc_Mutex* Lock) {
#ifdef DRIVER_ttc_mutex_init
    DRIVER_ttc_mutex_init(Lock);
#endif
}
         ttc_mutex_error_e ttc_mutex_lock_smart(ttc_mutex_smart_t* SmartMutex, Base_t TimeOut, void (*Owner)() ) {
    Assert_MUTEXES(SmartMutex, ec_InvalidArgument);

#if TTC_SMART_MUTEX_HISTORY_SIZE > 0
    tml_AmountSmartLocks++;
    tml_LastSmartMutex[tml_Index_LastSmartMutex++] = SmartMutex;
    if (tml_Index_LastSmartMutex >= TTC_SMART_MUTEX_HISTORY_SIZE)
        tml_Index_LastSmartMutex = 0;

    BOOL MutexIsKnown = FALSE;
    for (int I = 0; (I < tml_Index_UniqueSmartMutex) && (I < TTC_SMART_MUTEX_HISTORY_SIZE); I++) {
        if (tml_UniqueSmartMutex[I] == SmartMutex) {
            MutexIsKnown = TRUE;
            break;
        }
    }
    if ( (!MutexIsKnown) && (tml_Index_UniqueSmartMutex < TTC_SMART_MUTEX_HISTORY_SIZE) )
        tml_UniqueSmartMutex[tml_Index_UniqueSmartMutex++] = SmartMutex;

#endif

    ttc_mutex_error_e Error = ttc_mutex_lock(SmartMutex->Mutex, TimeOut);
    if (Error == tme_OK)
        SmartMutex->Owner = Owner;

    return Error;
}
         ttc_mutex_error_e ttc_mutex_lock(ttc_Mutex* Mutex, Base_t TimeOut) {
    Assert_MUTEXES(Mutex != 0, ec_InvalidArgument);

#if TTC_MUTEX_TIMEOUT_USECS > 0
    if (TimeOut == -1)
        TimeOut = TTC_MUTEX_TIMEOUT_USECS;
#endif

    ttc_mutex_error_e LockError = tme_NotImplemented;

#ifdef DRIVER_ttc_mutex_lock
    LockError = DRIVER_ttc_mutex_lock(Mutex, TimeOut);
#endif

#if TTC_MUTEX_TIMEOUT_USECS > 0
    if (LockError == tme_TimeOut) {
         Assert(TimeOut != TTC_MUTEX_TIMEOUT_USECS, ec_TimeOut); // endless mutex timed out!
    }
#endif

    return LockError;
}
                      void ttc_mutex_wait() {
    ttc_task_yield();
}
         ttc_mutex_error_e ttc_mutex_lock_isr(ttc_Mutex* Mutex) {
    Assert_MUTEXES(Mutex != 0, ec_InvalidArgument);
    static signed portBASE_TYPE HigherPrioTaskWoken; HigherPrioTaskWoken = FALSE;

#ifdef DRIVER_ttc_mutex_lock_isr
    return DRIVER_ttc_mutex_lock_isr(Mutex);
#else
    return tme_NotImplemented;
#endif
}
         ttc_mutex_error_e ttc_mutex_unlock(ttc_Mutex* Mutex) {
    Assert_MUTEXES(Mutex != 0, ec_InvalidArgument);

#ifdef DRIVER_ttc_mutex_unlock
    return DRIVER_ttc_mutex_unlock(Mutex);
#else
    return tme_NotImplemented;
#endif

}
         ttc_mutex_error_e ttc_mutex_unlock_isr(ttc_Mutex* Mutex) {
    Assert_MUTEXES(Mutex != 0, ec_InvalidArgument);

#ifdef DRIVER_ttc_mutex_unlock_isr
    return DRIVER_ttc_mutex_unlock_isr(Mutex);
#else
    return tme_NotImplemented;
#endif
}

// private functions ********************************************

         ttc_mutex_error_e _ttc_mutex_unlock_smart(ttc_mutex_smart_t* SmartMutex) {
    Assert_MUTEXES(SmartMutex, ec_InvalidArgument);

    if (0) SmartMutex->Owner = NULL;

    return ttc_mutex_unlock(SmartMutex->Mutex);
}

//}functions
