#ifndef ttc_mutex_H
#define ttc_mutex_H

/*{ ttc_mutex.h ***********************************************
 * 
 * Written by Gregor Rebel 2010-2012
 *
 * Architecture independent support for inter task communication & synchronization.
 * 
 * Currently supported architecture/ schedulers: cm3, FreeRTOS
 * 
}*/

//{ includes

#include "ttc_basic.h"

#ifdef EXTENSION_500_ttc_mutex_cortexm3
#  include "cm3_mutex.h"
#endif

#ifdef EXTENSION_500_ttc_mutex_freertos
#  include "freertos_mutex.h"
#endif

#include "ttc_mutex_types.h"

//}includes
//{ Types & Structures ***************************************************

/** Types to be defined by low-level driver ******************************
 *
 * ttc_Mutex  datatype storing a mutex
 * E.g.: typedef u8_t ttc_Mutex
 *
 *
 */

// enable use of smart mutexes
#ifndef TTC_SMART_MUTEXES
  #define TTC_SMART_MUTEXES 1
#endif

// define amount
#if TTC_SMART_MUTEXES == 1

  // set amount of entries to be stored in history of smart mutex usage (debugging only)
  #ifndef TTC_SMART_MUTEX_HISTORY_SIZE
    #define TTC_SMART_MUTEX_HISTORY_SIZE 0
  #endif

#endif

// debuggable mutex that can store its owner
// This comes handy if you wanna know, who has locked a mutex before.
#if TTC_SMART_MUTEXES == 1
  typedef struct ttc_mutex_smart_s {
    ttc_Mutex* Mutex;
    void (*Owner)();
  } ttc_mutex_smart_t;
#else
  typedef ttc_Mutex* ttc_mutex_smart_t;
#endif

//}
#include "ttc_memory.h"

//{ Function prototypes **************************************************
//
// For each function, a DRIVER_() pendant must be defined by low-level driver to provide an implementation.

/** creates a new non-recursive mutex lock
 * @return                 =! NULL: mutex has been created successfully (use handle to operate on mutex)
 */
ttc_Mutex* ttc_mutex_create();
#ifndef DRIVER_ttc_mutex_create
#error Missing low-level implementation! // check description above for function to be implemented
#endif

/** initializes an existing non-recursive mutex lock to the unlocked state
 * @Lock                 storage to be initialized as mutex
 */
void ttc_mutex_init(ttc_Mutex* Lock);
#ifndef DRIVER_ttc_mutex_init
#error Missing low-level implementation! // check description above for function to be implemented
#endif

/** Tries to obtain lock on mutex within timeout
 * @param Mutex  handle of already created mutex
 * @param TimeOut  !=-1: amount of system ticks to wait for lock
 *                 ==-1: wait forever
 * @return         == tme_OK: Lock has been successfully obtained within TimeOut
 */
ttc_mutex_error_e ttc_mutex_lock(ttc_Mutex* Mutex, Base_t TimeOut);
#ifndef DRIVER_ttc_mutex_lock
#error Missing low-level implementation! // check description above for function to be implemented
#endif

/** Tries to obtain lock on mutex; this function will never block
 * @param Mutex                handle of already created mutex
 * @return         == tme_OK: Lock has been successfully obtained within TimeOut
 *                 == tme_HigherPriorityTaskWoken: - a task of higher priority has been interrupted => call ttc_task_yield() at end of your interrupt service routine!
 *                 >= tme_Error: Error condition
 */
ttc_mutex_error_e ttc_mutex_lock_isr(ttc_Mutex* Mutex);
#ifndef DRIVER_ttc_mutex_lock_isr
#error Missing low-level implementation! // check description above for function to be implemented
#endif

/** Releases an already obtained lock from mutex
 * @param Mutex  handle of already created smart mutex
 * @return         == tme_OK: Lock has been released successfully
 */
ttc_mutex_error_e ttc_mutex_unlock(ttc_Mutex* Mutex);
#ifndef DRIVER_ttc_mutex_unlock
#error Missing low-level implementation! // check description above for function to be implemented
#endif

/** Releases an already obtained lock from mutex (does never block)
 * @param Mutex  handle of already created mutex
 * @return         == tme_OK: Lock has been released successfully
 *                 == tme_HigherPriorityTaskWoken: call task ttc_task_yield_isr() before exiting your ISR!
 */
ttc_mutex_error_e ttc_mutex_unlock_isr(ttc_Mutex* Mutex);
#ifndef DRIVER_ttc_mutex_unlock_isr
#error Missing low-level implementation! // check description above for function to be implemented
#endif

/** creates a new non-recursive mutex lock
 *
 * Smart mutexes store the owner of a lock.
 * Whenever a mutex stays locked because of a programming error, then the last locking
 * function can be identified by taking a look into the smart lock structure.
 *
 * @return                 =! NULL: mutex has been created successfully (use handle to operate on mutex)
 */
ttc_mutex_smart_t* ttc_mutex_create_smart();

/** Tries to obtain lock on a smart mutex within timeout
 *
 * Smart mutexes store the owner of a lock.
 * Whenever a mutex stays locked because of a programming error, then the last locking
 * function can be identified by taking a look into the smart lock structure.
 *
 * @param Mutex    handle of already created mutex
 * @param TimeOut  amount of system ticks to wait for lock
 * @param Owner    reference to function which wants to obtain the lock
 * @return         == tme_OK: Lock has been successfully obtained within TimeOut
 */
ttc_mutex_error_e ttc_mutex_lock_smart(ttc_mutex_smart_t* SmartMutex, Base_t TimeOut, void (*Owner)() );

/** Releases an already obtained lock from mutex
 * @param Mutex  handle of already created smart mutex
 * @return       == tme_OK: Lock has been released successfully
 */
ttc_mutex_error_e _ttc_mutex_unlock_smart(ttc_mutex_smart_t* SmartMutex);
#define ttc_mutex_unlock_smart(SmartMutex) _ttc_mutex_unlock_smart(SmartMutex)

/** Releases an already obtained lock from mutex (does never block)
 * @param Mutex  handle of already created mutex
 * @return         == tme_OK: Lock has been released successfully
 *                   == tme_HigherPriorityTaskWoken: call task ttc_task_yield_isr() before exiting your ISR!
 */
#define ttc_mutex_unlock_isr_smart(SmartMutex) _ttc_mutex_unlock_smart(SmartMutex)

/** called while waiting for mutex lock. Place your breakpoint inside to debug hanging locks
  */
void ttc_mutex_wait();

//}Function prototypes

#endif
