#ifndef TTC_NETWORK_H
#define TTC_NETWORK_H

/*{ ttc_network.h ***************************************************

 This source is published under the GNU LESSER GENERAL PUBLIC LICENSE (LGPL).
 See file LEGAL.txt in this software package for details.

 Authors: <Gregor Rebel>


 Description:

 Basic universal network support.
 Creation of flexible, reusable memory blocks for network packets.

}*/
//{ Defines/ TypeDefs ****************************************************

// These shall be defined somewhere before

#ifndef TTC_ASSERT_NETWORK    // any previous definition set (Makefile)?
#define TTC_ASSERT_NETWORK 1  // string asserts are enabled by default
#endif
#if (TTC_ASSERT_NETWORK == 1)  // use Assert()s in NETWORK code (somewhat slower but alot easier to debug)
  #define Assert_Network(Condition, ErrorCode) Assert(Condition, ErrorCode)
#else  // us no Assert()s in NETWORK code (somewhat smaller + faster, but crashes are hard to debug)
  #define Assert_Network(Condition, ErrorCode)
#endif

//} Defines
//{ Includes *************************************************************

#include "ttc_memory.h"

//} Includes
//{ Structures/ Enums ****************************************************


typedef struct ttc_network_packet_s { // ttc_network_packet_t - universal network packet

  /* Network packets are allocated with extra space in front and after payload area.
  * This allows to add extra data before and after Payload[] without copying it into other buffers.
  * After successfull packet creation, the whole buffer can be transmitted by a single
  * DMA operation because Header + Payload + Footer builds one continuous memory block.
  *
  * After usage, a memory packet can be returned to its memory
  * pool by calling ttc_network_packet_release().
  * network packet memory layout (32-bit architectures)
  *
  * Offset             Fieldname                Size
  *   0                ttc_network_packet_t      20
  *   0                Hint                       1
  *   1                Hint2                      1
  *   2                Size                       2
  *   4                *releaseBuffer()           4
  *   8                *Header                    4
  *  12                *Payload                   4
  *  16                *Footer                    4
  *  20                Packet buffer            Size_Header + Size_Payload + Size_Footer
  *  20                Header[]                 Size_Header
  *  20 + Size_Header  Payload[]                Size_Payload
  *  20 + Size_Header  Footer[]                 Size_Footer
  *     + Size_Footer
  */

     u8_t Hint;         // hint for release function how to handle this block (speeds up releasing blocks from multiple pools via single function)
 // u16_t MaxSize:         this entry is omitted to save memory. One may use an Array that stores size for each Hint value.
     u8_t Hint2;        // second hint field (currently only used for 16-bit padding)
    s16_t PayloadSize;  // >0: current amount of bytes stored in Payload[]; <0:  string in Payload[] is zero terminated and -PayloadSize is its max size
     void (*releaseBuffer)(struct ttc_network_packet_s* Packet);  // function that will take back ownership of this block
    u8_t* Header;       // points to continous memory area before Payload where current header starts
    u8_t* Payload;      // points to continous memory area containing payload data of this packet
    u8_t* Footer;       // points to continous memory area after Payload where current footer starts
} __attribute__((__packed__)) ttc_network_packet_t;

typedef struct ttc_network_virtual_packet_s { // ttc_network_virtual_packet_t - universal network packet that uses data of an exisiting memory block

  // Virtual network packets make use of an existing memory block to store the payload.
  // The idea is that data from a big buffer or constant data from flash must not be copied
  // but instead just being referred to.
  //
  // After usage, a memory packet can be returned to its memory
  // pool by calling ttc_network_block_packet_release().
  // This will also release the referenced memory block.

    u8_t Hint;                  // hint for release function how to handle this block (speeds up releasing blocks from multiple pools via single function)
// u16_t MaxSize:                  this entry is omitted to save memory. One may use an Array that stores size for each Hint value.
    u8_t Reserved;              // pad to 16-bit boundary
   u16_t PayloadSize;           // current amount of bytes stored in Payload[]
    void (*releaseBuffer)(struct ttc_network_virtual_packet_s* Packet);  // function that will take back ownership of this block
   u8_t* Header;                // points to continous memory area before Payload where current header starts
   u8_t* Payload;               // points to continous memory area containing payload data of this packet (inside Block->Buffer[])
   u8_t* Footer;                // points to continous memory area after Payload where current footer starts
   ttc_memory_block_t* Block;   // points to memory block that stores Payload
} __attribute__((__packed__)) ttc_network_virtual_packet_t;

//} Structures/ Enums
//{ Function prototypes **************************************************

/* allocates a memory block that can store given amount of bytes + some extra data for sending data over a network.
 *
 * Network packets are mostly like memory blocks as described in ttc_memory.h.
 * They can store a limited amount of payload data + some extra data before and after the payload.
 * When a network packet is created for transmission, it normally is processed in several layers of
 * a network stack. Each layer adds some extra data before the payload (called header) and after the
 * payload (called footer). When the packet build is complete, it is handed over to the hardware
 * to transmit it. For this final stage, it is fastest if the final packet including all of its
 * headers and footers builds one contiguos block in memory.
 * A smart network stack acvoids to copy data from the payload area as much as possible.
 * ttc_network_packet_t allows to reserve extra space in before and after Payload[].
 * When a ttc_network_packet_t enters the trans
 *
 * Note: All ttc_malloc() functions will call Assert_Halt_EC(ec_OutOfMemory) if allocation fails.
 *
 * @param Size_Payload   AmountBytes of bytes to be allocated for Payload[]
 * @param Size_Header    AmountBytes of bytes to be extra allocated for packet header (before Payload[])
 * @param Size_Footer    AmountBytes of bytes to be extra allocated for packet footer (after Payload[])
 * @param Hint           argument passed to releaseBuffer() to allow fast release of buffers from different pools by single function
 * @param releaseBuffer  pointer to function that will take back ownership of block
 * @return               pointer to newly allocated packet
 */
ttc_network_packet_t* ttc_network_alloc_packet(u16_t Size_Payload, u8_t Size_Header, u8_t Size_Footer, u8_t Hint, void (*releaseBuffer)(ttc_network_packet_t* Packet));

/* returns given packet to its creator by calling the stored release function
 *
 * @param Packet         instance of network packet allocated by ttc_network_alloc_packet() before
 */
void ttc_network_release_packet(ttc_network_packet_t* Packet);

/* resets pointer in given packet back to empty state
 *
 * @param Packet         instance of network packet allocated by ttc_network_alloc_packet() before
 */
void ttc_network_reset_packet(ttc_network_packet_t* Packet);

/* moves header pointer backward to allow prepending given amount of bytes
 *
 * Note: Will Assert if allocated header is exceeded!
 *
 * @param Packet     instance of network packet allocated by ttc_network_alloc_packet() before
 * @param Size       amount of extra bytes to reserve for additional header
 */
void ttc_network_add_header(ttc_network_packet_t* Packet, u8_t Size);

/* moves footer pointer forward to allow appending given amount of bytes
 *
 * Note: Will Assert if allocated footer is exceeded!
 *
 * @param Packet        instance of network packet allocated by ttc_network_alloc_packet() before
 * @param Size          amount of extra bytes to reserve for additional header
 * @param PacketMaxSize = Size_Payload + Size_Header + Size_Footer as given to ttc_network_alloc_packet() before
 */
void ttc_network_add_footer(ttc_network_packet_t* Packet, u8_t Size);

/* moves header pointer forward to remove given amount of bytes from header
 *
 * Note: Will Assert if allocated header is exceeded!
 *
 * @param Packet     instance of network packet allocated by ttc_network_alloc_packet() before
 * @param Size       amount bytes to remove from header
 */
void ttc_network_del_header(ttc_network_packet_t* Packet, u8_t Size);

/* moves footer pointer backward to remove given amount of bytes from footer
 *
 * Note: Will Assert if allocated footer is exceeded!
 *
 * @param Packet     instance of network packet allocated by ttc_network_alloc_packet() before
 * @param Size       amount bytes to remove from footer
 */
void ttc_network_del_footer(ttc_network_packet_t* Packet, u8_t Size);


//} Function prototypes

#endif //TTC_NETWORK_H
