/*{ ttc_queue.c ***********************************************
 *
 * Written by Gregor Rebel 2010-2012
 *
 * Architecture independent support for inter task communication & synchronization.
 * 
 * Currently supported schedulers: FreeRTOS
}*/
#include "ttc_queue.h"

//{ functions *************************************************

           ttc_QueueHandle ttc_queue_create(Base_t AmountOfEntries, Base_t SizeOfEntries) {
    ttc_QueueHandle NewQueue = NULL;

#ifdef EXTENSION_300_scheduler_free_rtos
    NewQueue = (ttc_QueueHandle) xQueueCreate(AmountOfEntries, SizeOfEntries);
#endif

    Assert_QUEUES(NewQueue != NULL, ec_Malloc); // could not create queue (no implementation/ not egnough memory)
    ttc_task_check_stack(); //D

    return NewQueue;
}
         ttc_queue_error_e ttc_queue_push_back(ttc_QueueHandle Queue, const void* Entry, Base_t TimeOut) {
    Assert_QUEUES(Queue != 0, ec_InvalidArgument);
    Assert_QUEUES(Entry != 0, ec_InvalidArgument);

    ttc_task_check_stack(); //D
#ifdef EXTENSION_300_scheduler_free_rtos
    if (xQueueSend( (xQueueHandle) Queue, Entry, TimeOut ) == pdTRUE)
        return tqe_OK;
    else
        return tqe_TimeOut;
#endif

    return tqe_NotImplemented;
}
         ttc_queue_error_e ttc_queue_push_back_isr(ttc_QueueHandle Queue, const void* Entry) {
    Assert_QUEUES(Queue != 0, ec_InvalidArgument);
    Assert_QUEUES(Entry != 0, ec_InvalidArgument);

#ifdef EXTENSION_300_scheduler_free_rtos
    portBASE_TYPE xHigherPriorityTaskWoken = pdFALSE;
    signed portBASE_TYPE Status = xQueueSendToBackFromISR( (xQueueHandle) Queue, Entry, &xHigherPriorityTaskWoken);

    if (xHigherPriorityTaskWoken == pdFALSE) {
        if (Status == pdPASS)
            return tqe_OK;
        else
            return tqe_QueuePushFailed;
    }
    else
        return tqe_HigherPriorityTaskWoken;
#endif

    return tqe_NotImplemented;
}
         ttc_queue_error_e ttc_queue_pop_front(ttc_QueueHandle Queue, void* Entry, Base_t TimeOut) {
    Assert_QUEUES(Queue != 0, ec_InvalidArgument);
    Assert_QUEUES(Entry != 0, ec_InvalidArgument);

#ifdef EXTENSION_300_scheduler_free_rtos
    if ( xQueueReceive( (xQueueHandle) Queue, Entry, TimeOut ) )
        return tqe_OK;
    else
        return tqe_TimeOut;
#endif

    return tqe_NotImplemented;
}
         ttc_queue_error_e ttc_queue_pop_front_isr(ttc_QueueHandle Queue, void* Entry) {
    Assert_QUEUES(Queue != 0, ec_InvalidArgument);
    Assert_QUEUES(Entry != 0, ec_InvalidArgument);

#ifdef EXTENSION_300_scheduler_free_rtos
    signed portBASE_TYPE pxTaskWoken = 0;
    if ( xQueueReceiveFromISR( (xQueueHandle) Queue, Entry, &pxTaskWoken) ) {
        if (pxTaskWoken)
            return tqe_HigherPriorityTaskWoken;
        return tqe_OK;
    }
    else
        return tqe_TimeOut;
#endif

    return tqe_NotImplemented;
}
                    Base_t ttc_queue_amount_waiting(ttc_QueueHandle Queue) {
    Assert_QUEUES(Queue != 0, ec_InvalidArgument);

#ifdef EXTENSION_300_scheduler_free_rtos
    return uxQueueMessagesWaiting( (xQueueHandle) Queue );
#endif

    return 0;
}
       ttc_SemaphoreHandle ttc_semaphore_create(u8_t MaxCount, u8_t InitialCount) {

#ifdef EXTENSION_300_scheduler_free_rtos
        return xQueueCreateCountingSemaphore(MaxCount, InitialCount);
#endif

    return NULL;
}
         ttc_queue_error_e ttc_semaphore_take(ttc_SemaphoreHandle Semaphore, Base_t TimeOut) {
    Assert_QUEUES(Semaphore != 0, ec_InvalidArgument);

#ifdef EXTENSION_300_scheduler_free_rtos
    if (xSemaphoreTake(Semaphore, (portTickType) TimeOut))
        return tqe_OK;
    else
        return tqe_TimeOut;
#endif

    return tqe_NotImplemented;
}
         ttc_queue_error_e ttc_semaphore_give(ttc_SemaphoreHandle Semaphore) {
    Assert_QUEUES(Semaphore != 0, ec_InvalidArgument);

#ifdef EXTENSION_300_scheduler_free_rtos
    if ( xSemaphoreGive(Semaphore) )
        return tqe_OK;
    else
        return tqe_UNKNOWN;
#endif

    return tqe_NotImplemented;
}
         ttc_queue_error_e ttc_semaphore_give_isr(ttc_SemaphoreHandle Semaphore) {
    Assert_QUEUES(Semaphore != 0, ec_InvalidArgument);

#ifdef EXTENSION_300_scheduler_free_rtos
    portBASE_TYPE xHigherPriorityTaskWoken = pdFALSE;
    xSemaphoreGiveFromISR(Semaphore, &xHigherPriorityTaskWoken);

    if (xHigherPriorityTaskWoken == pdFALSE)
        return tqe_OK;
    else
        return tqe_HigherPriorityTaskWoken;
#endif

    return tqe_NotImplemented;
}
   ttc_QueueHandle_Bytes_t ttc_queue_byte_create(u16_t AmountOfEntries) {
    Assert_QUEUES(AmountOfEntries, ec_InvalidArgument); // need at least size 1

    ttc_SimpleQueueBytes_t* Queue = ttc_memory_alloc(AmountOfEntries + sizeof(ttc_SimpleQueueBytes_t));
    Queue->Size  = AmountOfEntries;
    Queue->Start = 0;
    Queue->End   = 0;
    Queue->Lock = ttc_mutex_create();

    return Queue;
}
                    Base_t ttc_queue_byte_get_amount(ttc_QueueHandle_Bytes_t QueueHandle) {
    ttc_SimpleQueueBytes_t* Queue = (ttc_SimpleQueueBytes_t*) QueueHandle;
    Assert_QUEUES(Queue, ec_InvalidArgument);

    if (Queue->End < Queue->Start)
        return Queue->End + Queue->Size - Queue->Start;

    return Queue->End - Queue->Start;
}
         ttc_queue_error_e ttc_queue_byte_push_back(ttc_QueueHandle_Bytes_t QueueHandle, u8_t Data) {
    ttc_SimpleQueueBytes_t* Queue = (ttc_SimpleQueueBytes_t*) QueueHandle;
    Assert_QUEUES(Queue, ec_InvalidArgument);
    ttc_mutex_lock(Queue->Lock, -1);
    ttc_queue_error_e Error = _ttc_queue_byte_push_back(Queue, Data);
    ttc_mutex_unlock(Queue->Lock);

    return Error;
}
         ttc_queue_error_e ttc_queue_byte_push_back_isr(ttc_QueueHandle_Bytes_t QueueHandle, u8_t Data) {
    // variables may be static because we don't expect concurrency inside ISR
    static ttc_queue_error_e LockError;
    static ttc_queue_error_e QueueError;

    ttc_SimpleQueueBytes_t* Queue = (ttc_SimpleQueueBytes_t*) QueueHandle;
    Assert_QUEUES(Queue, ec_InvalidArgument);
    LockError = ttc_mutex_lock_isr(Queue->Lock);
    if (!LockError) {
        QueueError = _ttc_queue_byte_push_back(Queue, Data);
    }
    else { // queue is currently locked by a task
        if (!QueueError)
            QueueError = LockError; // maybe == tqe_HigherPriorityTaskWoken
    }

    ttc_mutex_unlock(Queue->Lock);


    return QueueError;
}
         ttc_queue_error_e ttc_queue_byte_pull_front(ttc_QueueHandle_Bytes_t QueueHandle, u8_t* Data) {
    ttc_SimpleQueueBytes_t* Queue = (ttc_SimpleQueueBytes_t*) QueueHandle;
    Assert_QUEUES(Queue, ec_InvalidArgument);
    Assert_QUEUES(Data,  ec_InvalidArgument);

    ttc_mutex_lock(Queue->Lock, -1);
    ttc_queue_error_e Error = _ttc_queue_byte_pull_front(Queue, Data);
    ttc_mutex_unlock(Queue->Lock);

    return Error;
}
         ttc_queue_error_e ttc_queue_byte_pull_front_isr(ttc_QueueHandle_Bytes_t QueueHandle, u8_t* Data) {
    // variables may be static because we don't expect concurrency inside ISR
    static ttc_queue_error_e LockError;
    static ttc_queue_error_e QueueError;

    ttc_SimpleQueueBytes_t* Queue = (ttc_SimpleQueueBytes_t*) QueueHandle;
    Assert_QUEUES(Queue, ec_InvalidArgument);
    Assert_QUEUES(Data,  ec_InvalidArgument);

    LockError = ttc_mutex_lock_isr(Queue->Lock);
    if (!LockError) {
        QueueError = _ttc_queue_byte_pull_front(Queue, Data);
    }
    else { // queue is currently locked by a task

        if (!QueueError)
            QueueError = LockError;
    }

    ttc_mutex_unlock(Queue->Lock);


    return QueueError;
}
ttc_QueueHandle_Pointers_t ttc_queue_pointer_create(u16_t AmountOfEntries) {
    Assert_QUEUES(AmountOfEntries, ec_InvalidArgument); // need at least size 1

    ttc_SimpleQueuePointers_t* Queue = ttc_memory_alloc(sizeof(void*) * AmountOfEntries + sizeof(ttc_SimpleQueuePointers_t));
    Queue->Size  = AmountOfEntries;
    Queue->Start = 0;
    Queue->End   = 0;
    Queue->Lock  = ttc_mutex_create();

    return Queue;
}
                    Base_t ttc_queue_pointer_get_amount(ttc_QueueHandle_Pointers_t QueueHandle) {
    ttc_SimpleQueuePointers_t* Queue = (ttc_SimpleQueuePointers_t*) QueueHandle;
    Assert_QUEUES(Queue, ec_InvalidArgument);

    if (Queue->End < Queue->Start)
        return Queue->End + Queue->Size - Queue->Start;

    return Queue->End - Queue->Start;
}
         ttc_queue_error_e ttc_queue_pointer_push_back(ttc_QueueHandle_Pointers_t QueueHandle, void* Pointer) {
    ttc_SimpleQueuePointers_t* Queue = (ttc_SimpleQueuePointers_t*) QueueHandle;
    Assert_QUEUES(Queue, ec_InvalidArgument);

    if (1) ttc_mutex_lock(Queue->Lock, -1);

    Base_t NewEnd = Queue->End + 1;
    if (NewEnd >= Queue->Size)
        NewEnd = 0;                 // wrapping around at end of array

    if (Queue->Start == NewEnd) {
        if (1) ttc_mutex_unlock(Queue->Lock);
        return tqe_QueuePushFailed; // queue is full
    }

    Queue->Data[Queue->End] = Pointer; // storing at old end index
    Queue->End = NewEnd;

    if (1) ttc_mutex_unlock(Queue->Lock);
    return tqe_OK;
}
         ttc_queue_error_e ttc_queue_pointer_pull_front(ttc_QueueHandle_Pointers_t QueueHandle, void** Pointer) {
    ttc_SimpleQueuePointers_t* Queue = (ttc_SimpleQueuePointers_t*) QueueHandle;
    Assert_QUEUES(Queue,   ec_InvalidArgument);
    Assert_QUEUES(Pointer, ec_InvalidArgument);

    if (1) ttc_mutex_lock(Queue->Lock, -1);

    u16_t Start = Queue->Start;
    Assert_QUEUES(Start < Queue->Size,  ec_UNKNOWN);

    if (Start == Queue->End) {   // queue is empty
        *Pointer = NULL;
        if (1) ttc_mutex_unlock(Queue->Lock);
        return tqe_QueueIsEmpty;
    }
    void* Address = Queue->Data[Start++];
    *Pointer = Address;

    if (Start == Queue->Size)
        Start = 0;              // wrapping around at end of array
    Queue->Start = Start;

    if (1) ttc_mutex_unlock(Queue->Lock);
    return tqe_OK;
}

// private functions ********************************************

         ttc_queue_error_e _ttc_queue_byte_push_back(ttc_SimpleQueueBytes_t* Queue, u8_t Data) {

    u16_t NewEnd = Queue->End + 1;
    if (NewEnd >= Queue->Size)
        NewEnd = 0;

    if (Queue->Start == NewEnd) {
        ttc_mutex_unlock(Queue->Lock);
        return tqe_QueuePushFailed; // queue is full
    }

    Queue->Data[Queue->End] = Data; // storing at old end index
    Queue->End = NewEnd;

    return tqe_OK;
}
         ttc_queue_error_e _ttc_queue_byte_pull_front(ttc_SimpleQueueBytes_t* Queue, u8_t* Data) {
    u16_t Start = Queue->Start;
    Assert_QUEUES(Start < Queue->Size,  ec_UNKNOWN);

    if (Start == Queue->End) {      // queue is empty
        ttc_mutex_unlock(Queue->Lock);
        return tqe_QueueIsEmpty;
    }

    u8_t Byte = Queue->Data[Start++];
    *Data = Byte;

    if (Start == Queue->Size)
        Start = 0;                 // wrapping around at end of array
    Queue->Start = Start;

    return tqe_OK;
}

//}functions
