#ifndef ttc_task_H
#define ttc_task_H

/*{ ttc_task.h ***********************************************
 
 * Written by Gregor Rebel 2010-2012
 *
 * Architecture independent support for multitasking.
 * 
 * Currently implemented architectures: stm32 + FreeRTOS
 * 
}*/

//{ includes

#include "ttc_basic.h"
#include "ttc_gpio.h"

#ifdef EXTENSION_300_scheduler_free_rtos
  #include <string.h>
  #include "stddef.h"
  #include "FreeRTOS.h"
  #include "task.h"
  #include "portmacro.h"

  #define TTC_TASK_MINIMAL_STACK_SIZE configMINIMAL_STACK_SIZE
#endif

//}includes
//{ constants

#ifndef TTC_ASSERT_TASK    // any previous definition set (Makefile)?
#define TTC_ASSERT_TASK 1  // string asserts are enabled by default
#endif
#if (TTC_ASSERT_TASK == 1)  // use Assert()s in TASK code (somewhat slower but alot easier to debug)
  #define Assert_Task(Condition, ErrorCode) Assert(Condition, ErrorCode)
#else  // us no Assert()s in TASK code (somewhat smaller + faster, but crashes are hard to debug)
  #define Assert_Task(Condition, ErrorCode)
#endif

#ifndef configTICK_RATE_HZ
  #define configTICK_RATE_HZ 0
#endif

#ifndef TTC_TASK_MINIMAL_STACK_SIZE
  #define TTC_TASK_MINIMAL_STACK_SIZE 128
#endif

//}

#ifdef EXTENSION_300_scheduler_free_rtos
  // System Tick Counter provided by FreeRTOS
  extern volatile portTickType xTickCount;
#endif

typedef void* TaskHandle_t;

//{ Function prototypes **************************************************

/** produces software PWM usable to calibrate uSleep()
 * Idea: Connect oscilloscope to pin PA1 (pin 15 on STM32F103) an check for correct period (1,1ms) and duty cycle.
 *        Adjust uSleep() accordingly, recompile, reflash and retest.
 */
void programCalibrateUSleep();

/** Set task to sleep for some amount of time
 * @param Delay    delay time (msecs)
*/
void ttc_task_msleep(u32_t Delay);
void mSleep(u32_t Delay);

/** set task to sleep for some amount of time
 * @param Delay    delay time (usecs)
 */
void ttc_task_usleep(u32_t Delay);
void uSleep(u32_t Delay);

/** spawns a function as separate task
 * @param TaskFunction  function to start as task (must never return!)
 * @param TaskName      used for debugging purposes only
 * @param StackSize     each task has its own size; choose wisely
 * @param Argument      passed as argument to TaskFunction()
 * @param Priority      tasks with higher priority get more cpu time
 * @param Handle        =! NULL: stores task handle in *Handle; A task handle can be used to destroy the task later.
 * @return              == TRUE: task was created successfully
 */
BOOL ttc_task_create( void TaskFunction(void*), const char* TaskName, u16_t StackSize, void* Argument, u8_t Priority, TaskHandle_t* Handle);


/** Switches scheduler to next process providing a short, undetermined sleep
 *
 * Some applications require a busy waiting for resources like:
 *   while (...);
 * During this loop, the cpu is completely busy with this loop. This blocks or at least slows down other tasks.
 * Even if this task can be interrupted by the task scheduler, it fills its time slice with checking its condition (...).
 *
 * A more multitask-friendly version is to yield cpu time to other processes while waiting:
 *   while (...) ttc_task_yield();
 * For every checked condition (...), every task gets one slice of cpu time. The processing overhead is minimal.
 *
 */
void ttc_task_yield();

/** similar to ttc_task_yield() for use in interrupt service routines only
  */
void ttc_task_yield_isr();

/** Temporarily disables task scheduler*/
#ifdef TTC_ASSERT_TASK
#define ttc_task_begin_criticalsection(OWNER) _ttc_task_begin_criticalsection2( (void (*)()) OWNER)
#else
#define ttc_task_begin_criticalsection(OWNER) _ttc_task_begin_criticalsection()
#endif

void _ttc_task_begin_criticalsection();

void _ttc_task_begin_criticalsection2(void (*Owner)());

/** Reenables task scheduler */
void ttc_task_end_criticalsection();

/** returns current system clock value
 *
 * Note: Check configuration of your scheduler to achieve real world usecs! (E.g. FreeRTOSConfig.h:configTICK_RATE_HZ)
 *
 * @return amount of usecs since start of system/ scheduler
 */
Base_t ttc_task_get_tick_count();

/** checks if stack of current task or any other task has been overflowed in the past.
 * @param != 0: handle of task which stack should be inspected; ==0: inspect current task
 * @return minimum seen stack bytes free so far
 */
Base_signed_t ttc_task_check_stack();

/** starts multitasking scheduler (will not return if scheduler is available)
 */
void ttc_task_start_scheduler();

/** resumes a suspended task
  *
  * Note: Do not call this function from an interrupt service routine!
  *
  * @param Handle task identifier as returned by ttc_task_create
  */
void ttc_task_resume(TaskHandle_t Handle);

/** resumes a suspended task from an interrupt service routine
  *
  * @param   Handle task identifier as returned by ttc_task_create
  * @return  == TRUE: you must call ttc_task_yield_isr() at end of your isr!
  */
BOOL ttc_task_resume_isr(TaskHandle_t Handle);

/** puts a task to sleep.
  *
  * @param Handle != NULL: task identifier as returned by ttc_task_create
  *               == NULL: put current task to sleep
  */
void ttc_task_suspend(TaskHandle_t Handle);

#ifdef EXTENSION_300_scheduler_free_rtos

/** called in case pvPortMalloc() fails */
void vApplicationMallocFailedHook(void);

void vApplicationStackOverflowHook( xTaskHandle pxTask, signed char *pcTaskName );

#endif
//}Function prototypes

#endif
