/*{ regression_memory_pools.c ************************************************
 
  regression_start_memory_pools() spawns thread taskBrightness()

  taskBrightness()
      Defines brightness of led as Ton / (Ton + Toff)
      spawns thread taskLEDs()
      while (1) {
        while (Ton < 100) { Ton++; Toff--; }
        while (Ton >   0) { Ton--; Toff++; }
      }

  taskLEDs()
      while (1) {
        switch on LED
        wait for Ton
        switch off LED
        wait for Toff
      }

}*/

#include "regression_memory_pools.h"

// read-only configuration of current display (width, height, ...)
ttc_gfx_generic_t* DisplayConfig = NULL;

//{ Function definitions *************************************************
//{ RMA-Implementation ****************************************************

#define createPool(SIZE)    ttc_memory_pool_create(SIZE,100)
#define mallocBlock(POOL)   ttc_memory_pool_alloc(POOL)
#define releaseBlock(BLOCK) ttc_memory_pool_release(BLOCK)
#define mutex_create ttc_mutex_create
#define mutex_lock   ttc_mutex_lock
#define mutex_unlock ttc_mutex_unlock
#define MutexHandle  ttc_MutexHandle
#define QueueHandle  ttc_QueueHandle_Pointers_t
#define enqueue(Q,P) ttc_queue_pointer_push_back(Q, P)
#define dequeue(Q,P) ttc_queue_pointer_pull_front(Q, (void**) P)
#define random(MAX)  ttc_rand() % MAX
#define task_create(FUNC, ARG) ttc_task_create(FUNC, "", 256, (void*) ARG, 0, NULL)
#define LED1         PIN_PE0
#define LED2         PIN_PE1

// structures

// global variables
static ttc_memory_pool_t* Pool100 = NULL;

//} RMA-Implementation
//{ Global Variables *****************************************************

static ttc_memory_pool_t* PoolPrintJobs;
void taskProducer(void* Arg);
void taskConsumer(void* Arg);
u8_t DisplayY = 2;
QueueHandle PrintQueue;
QueueHandle BlockQueue;

//} Global Variables

void regression_init_memory_pools() {

    u8_t MaxDisplay = ttc_gfx_get_max_display_index();
    Assert(MaxDisplay > 0, ec_DeviceNotFound); // no displays defined by board makefile!
    DisplayConfig = ttc_gfx_get_configuration(1);
    Assert(DisplayConfig, ec_DeviceNotFound);

    // set color to use for first screen clear
    ttc_gfx_set_color_bg24(GFX_COLOR24_AQUA);
    ttc_gfx_init(1);

    // initialize gfx for use with printf()
    ttc_gfx_stdout_set(1);
    ttc_string_register_stdout(ttc_gfx_stdout_print_block, ttc_gfx_stdout_print_string);
}
void regression_start_memory_pools() {
    regression_init_memory_pools();

    // TestRMA
    ttc_gpio_init(LED1, tgm_output_push_pull);

    PoolPrintJobs = createPool(sizeof(printJob_t));
    PrintQueue = ttc_queue_pointer_create(100);
    BlockQueue = ttc_queue_pointer_create(100);

    Pool100 = createPool(100);

    const char* Title = "Test RMA";
    ttc_gfx_text_cursor_set(ttc_gfx_text_center(Title, -1), 0);
    ttc_gfx_print(Title, -1);

    Base_t SleepMax = 88;
    for (u8_t Amount = 8; Amount > 0; Amount--)
        task_create(taskProducer, SleepMax);

    SleepMax = 20;
    for (u8_t Amount = 2; Amount > 0; Amount--)
        task_create(taskConsumer, SleepMax);

    task_create(taskPrint, NULL);

    mSleep(400); // wait for tasks to settle

    //? printAt(0, DisplayConfig->TextRows - 1, "used: %iB", AmountAllocatedBytes);
    printAt(0, 1, "Pa%i ", Pool100->Statistics.AmountAllocated);
    printAt(5, 1, "PF%i ", Pool100->Statistics.AmountFreeMaximum);
    printAt(10, 1, "Pf%i ", 0);
    while (1) {
        //? printAt(6, DisplayConfig->TextRows - 1, "%iB", ((Base_t) NextFree) - ((Base_t) Heap));
        printAt( 2, 1, "%i ", Pool100->Statistics.AmountAllocated);
        printAt( 7, 1, "%i ", Pool100->Statistics.AmountFreeMaximum);
        if (Pool100->Statistics.AmountUseTotal)
            printAt(12, 1, "%i ", Pool100->Statistics.AmountFreeTotal / Pool100->Statistics.AmountUseTotal);
        mSleep(1000);
    }
}

void taskProducer(void* Arg) {
    Base_t MaxSleepTime = (Base_t) Arg;
    Base_t SleepTotal = 0;
    u8_t Y = DisplayY++;
    printAt(0, Y, "P>", 0);
    Base_t TimeSecs = ttc_task_get_tick_count() / 1000000;
    Base_t BlocksProcessed = 1;

    //while (1)
    mSleep(700);

    ttc_memory_from_pool_t* Block = NULL;
    printAt(3, Y, "T %i", SleepTotal / BlocksProcessed);

    while (1) {
        Base_t Seconds = ttc_task_get_tick_count() / 1000000;
        if (TimeSecs != Seconds) { // one screen update every second
            TimeSecs = Seconds;
            if (Seconds & 1)
                printAt(4, Y, "|%i", SleepTotal / BlocksProcessed);
            else
                printAt(4, Y, "-%i", SleepTotal / BlocksProcessed);
        }

        ttc_gpio_set(LED1);
        Block = mallocBlock(Pool100);
        ttc_gpio_clr(LED1);

        if (Block) {
            BlocksProcessed++;
            unsigned char* Data = (unsigned char*) Block;
            (void) Data;

            // simulate random time required to fill Data[]
            Base_t SleepTime = 1 + random(MaxSleepTime);
            SleepTotal += SleepTime;
            mSleep(SleepTime);

            while (enqueue(BlockQueue, Block) != tqe_OK)
                ttc_task_yield();
        }
        else
            ttc_task_yield();
    }
}
void taskConsumer(void* Arg) {
    Base_t MaxSleepTime = (Base_t) Arg;
    Base_t SleepTotal = 0;
    ttc_memory_from_pool_t* Block = NULL;
    u8_t Y = DisplayY++;
    printAt(0, Y, "C>", 0);

    //while (1)
    mSleep(500);

    Base_t TimeSecs = ttc_task_get_tick_count() / 1000000;
    Base_t BlocksProcessed = 0;
    printAt(3, Y, "T %i", BlocksProcessed);

    while (1) {
        Base_t Seconds = ttc_task_get_tick_count() / 1000000;
        if (TimeSecs != Seconds) { // one screen update every second
            TimeSecs = Seconds;
            if (BlocksProcessed) {
                if (Seconds & 1)
                    printAt(4, Y, "|%i", SleepTotal / BlocksProcessed);
                else
                    printAt(4, Y, "-%i", SleepTotal / BlocksProcessed);
            }
        }

        if ((dequeue(BlockQueue, &Block) == tqe_OK)) {
            Assert(Block, ec_NULL);
            unsigned char* Data = (unsigned char*) Block;
            (void) Data;

            // simulate random processing time of Data[]
            Base_t SleepTime = 1 + random(MaxSleepTime);
            SleepTotal += SleepTime;
            mSleep(SleepTime);
            BlocksProcessed++;

            ttc_gpio_set(LED2);
            releaseBlock(Block);
            ttc_gpio_clr(LED2);
        }
        else
            ttc_task_yield();
    }
}
void printAt(u8_t X, u8_t Y, const char* Format, u32_t Value) {
    ttc_memory_from_pool_t* Block = mallocBlock(PoolPrintJobs);
    Assert(Value < 10000, ec_Debug); //D

    printJob_t* PrintJob = (printJob_t*) Block;
    Assert(PrintJob, ec_NULL);
    Assert( ( ((Base_t) PrintJob) > 0x20000000) && ( ((Base_t) PrintJob) < 0x20010000), ec_UNKNOWN);
    ttc_memory_pool_t* Pool = (Block-1)->Pool;
    Assert(Pool, ec_NULL);
    Assert( ( ((Base_t) Pool) > 0x20000000) && ( ((Base_t) Pool) < 0x20010000), ec_UNKNOWN);

    PrintJob->X      = X;
    PrintJob->Y      = Y;
    PrintJob->Format = Format;
    PrintJob->Value  = Value;

    enqueue(PrintQueue, Block);
}
void taskPrint(void* Arg) {
    (void) Arg;

    ttc_memory_from_pool_t* MemoryBlock;
    printJob_t* PrintJob;
    static u8_t Temp[100]; // there can be only one taskPrint() instance

    while (1) {
        if (dequeue(PrintQueue, &MemoryBlock) == tqe_OK) {
            PrintJob = (printJob_t*) MemoryBlock;
            Assert(PrintJob, ec_NULL);
            Assert( ( ((Base_t) PrintJob) > 0x20000000) && ( ((Base_t) PrintJob) < 0x20010000), ec_UNKNOWN);
            if ( (PrintJob->X != -1) && (PrintJob->Y) )
                ttc_gfx_text_cursor_set(PrintJob->X, PrintJob->Y);
            Assert(PrintJob->Value < 100000, ec_Debug); //D
            ttc_string_snprintf(Temp, 100, PrintJob->Format, PrintJob->Value);
            ttc_gfx_print_solid( (char*) Temp, 100);
            releaseBlock(MemoryBlock);
        }
        else
            ttc_task_yield();
    }
}

//} Function definitions
