#!/bin/bash

#
#  Install script for Communication Application Peripheral Library (CPAL) 
#  The CPAL library has been released by ST Microelectronics Corp. as
#  a high-level interface for the different standard peripheral libaries
#  for STM32 microcontrollers
#
#  Script written by Gregor Rebel 2010-2012.
# 
#  Feel free do distribute and change to your own needs!
#
# Tested successfully on:
# openSuSE 11.4 x86
# XUbuntu 11.10 x64

source scripts/installFuncs.sh

# ensure that $PATH contains no . (brings find into trouble)
#export PATH=`perl -e "my @A=split(':', '$PATH');  print join(':', grep { \\\$_ ne '.' } @A );"`
#export PATH=`perl -e "my @A=split(':', '$PATH');  print join(':', grep { substr(\\\$_,0,1) ne '.' } @A );"`

if [ "$USER" == "root" ]; then
  echo "$0: ERROR this script should not been run as user root!"
  exit 10
fi

INSTALLPATH="270_CPAL"
dir "$INSTALLPATH"
cd "$INSTALLPATH"

#{ IMPLEMENT YOUR INSTALL SCRIPT BELOW -----------------------------------------------

if [ ! -e OK.Docs ];       then #{ download documentation

  Get_Fails=""
  
  File="STM32F10x_Standard_Peripherals_Library.html"
  getFile http://www.mikrocontroller.net/articles/ STM32F10x_Standard_Peripherals_Library $File
  addDocumentationFile $File STM
  
  File="UM1029-Communication_peripheral_application_library_CPAL.pdf"
  getFile http://www.st.com/internet/com/TECHNICAL_RESOURCES/TECHNICAL_LITERATURE/USER_MANUAL/ CD00291090.pdf $File 
  addDocumentationFile $File STM
  #X ln -s ../../Documentation/STM/UM1029-Communication_peripheral_application_library_CPAL.pdf .
  if [ "$Get_Fails" == "" ]; then
    touch OK.Docs
  else
    echo "$0 - ERROR: missing files: $Get_Fails"
  fi
fi #}
if [ ! -e OK.Install ];    then #{ download + install software
  if [ ! -e OK.Packages ]; then #{ install packages
    installPackage libgmp3-dev
    installPackage libmpfr-dev
    installPackage xchm
    installPackage chmlib
    installPackage chmsee
    installPackage kchmviewer
    touch OK.Packages
  fi #}
  if [ -e OK.Packages ];   then #{ install software
    echo "installing in $INSTALLPATH ..."
    InstallDir=`pwd`
    
    getFile http://www.st.com/internet/com/SOFTWARE_RESOURCES/SW_COMPONENT/FIRMWARE/ stm32_i2c_cpal.zip stm32_i2c_cpal.zip
    unZIP "UNKNOWN" stm32_i2c_cpal.zip

    Current=`find ./ -name STM32_I2C_CPAL* -type d | sort | tail -n 1`
    if [ "$Current" != "" ]; then #{ create link current
      rm 2>/dev/null current
      ln -sv $Current current
    fi #}

    if [ -e current/Release_Notes.html ]; then
      echo "" >OK.Install
    fi
    #cd ..
  else
    echo "$0 - ERROR: Packages were not installed correctly!"
  fi #}OK.Packages
fi #}
if [   -e OK.Install ];    then #{ create makefiles and activate scripts
  Name="${INSTALLPATH}"
  
  if [ "" == "DISABLED" ]; then #{ Currently not used
  createExtensionMakefileHead ${Name}      #{ makefile CPAL
  cat <<END_OF_MAKEFILE >>../extensions/makefile.$Name 

# activate certain parts of source-code
COMPILE_OPTS += -DEXTENSION_${Name}=1

INCLUDE_DIRS += -I additionals/$Name/

END_OF_MAKEFILE
  createExtensionMakefileTail ${Name} #}
  File="../extensions/activate.${Name}.sh" #{ activate CPAL
  createActivateScriptHead $File "$0" "Communication Peripheral Access Library is a superset of ST's standard peripheral libraries"
  cat <<END_OF_ACTIVATE >>$File
#!/bin/bash

rm 2>/dev/null \$Dir_ExtensionsActive/makefile.250_stm_std_peripherals_*
rm 2>/dev/null \$Dir_Additionals/250_stm_std_peripherals

OldPWD=\`pwd\`
cd "\$Dir_Additionals"
LinkScript="createLink.sh"

rm "\$LinkScript"
cd "\$OldPWD"

# create links into extensions.active/
createLink \$Dir_Extensions/makefile.$Name \$Dir_ExtensionsActive/makefile.$Name '' QUIET

END_OF_ACTIVATE
  createActivateScriptTail $File $Name
  #}
  fi #}
  
  addLine ../scripts/createLinks.sh "rm 2>/dev/null $INSTALLPATH;  ln -sv \$Source/ToolChain_STM32/ToolChain/$INSTALLPATH/current  $INSTALLPATH"
  OldPWD="`pwd`"
  Dir_CoreSupport=`../scripts/findDirectory.pl core_cm3.h current/`
  cd current/Libraries/

    #{ create macro assert_param in stm32f10x.h to fix a compilation issue ToDo: Needed anymore?
    
    File="CMSIS/CM3/DeviceSupport/ST/STM32F10x/stm32f10x.h"
    if [ -e ${File}_orig ]; then
        mv -f ${File}_orig $File 
    fi
    mv -f $File ${File}_orig
    echo "  creating macro assert_param in $File"
    Note="  Note: Manually remove assert_param from your stm32f10x_conf.h !"
    echo $Note
    addLine ../Notes.txt $Note
    addLine ../Notes.txt "StdPeripheralLibrary-> http://www.mikrocontroller.net/articles/STM32F10x_Standard_Peripherals_Library"
    
    cat >script.sed <<"END_OF_TEXT" #{
/#include <stdint.h>/a\
#ifdef  USE_FULL_ASSERT\n
\n
// This assert_param macro has been copied here from stm32f10x_conf.h to\n
// circumvent gcc compiler errors.\n
\n
/**\n
  * @brief  The assert_param macro is used for function's parameters check.\n
  * @param  expr: If expr is false, it calls assert_failed function\n
  *   which reports the name of the source file and the source\n
  *   line number of the call that failed.\n 
  *   If expr is true, it returns no value.\n
  * @retval None\n
  */\n
  #define assert_param(expr) ((expr) ? (void)0 : assert_failed((uint8_t *)__FILE__, __LINE__))\n
  void assert_failed(uint8_t* file, uint32_t line);\n
#else\n
  #define assert_param(expr) ((void)0)\n
#endif /* USE_FULL_ASSERT */\n

END_OF_TEXT
#}
    
    # concatenate all lines which contain \n (sed requires this)
    awk '/\\n/{printf "%s",$0;next}{print}' script.sed >script2.sed
    
    # insert lines after "#include <stdint.h>"
    sed -f script2.sed ${File}_orig >$File
    #} fix a bug
  
  SubDirs=`ls */ -d`
  for SubDir in $SubDirs; do #{ create link into additionals/ for each sub directory
    SubDir=`perl -e "print substr('${SubDir}', 0, -1);"` # remove trailing /
    echo "sub library: $SubDir"
    Link="${INSTALLPATH}_$SubDir"
    addLine $OldPWD/../scripts/createLinks.sh "rm 2>/dev/null $Link;  ln -sv $INSTALLPATH/Libraries/$SubDir  $Link"
  done #}
  
  SubDirs=`ls -d *_StdPeriph_Driver`
  for SubDir in $SubDirs; do #{ create makefile + activate script for each *_StdPeriph_Driver directory
    Name="250_CPAL_${SubDir}" # std peripheral uses different rank than CPAL!
    Name2="270_CPAL_${SubDir}"
    Architecture=`perl -e "print substr('${SubDir}', 0, 9);"`

    createExtensionMakefileHead ${Name} $OldPWD/../extensions/ #{      create makefile for std_peripheral_library
    cat <<END_OF_MAKEFILE >>$MakeFile

COMPILE_OPTS += -DEXTENSION_250_stm_std_peripherals=1

INCLUDE_DIRS += -I additionals/250_stm_std_peripherals/inc/ \\
                -I additionals/270_CPAL_CMSIS/CM3/CoreSupport/ \\
                -I additionals/270_CPAL_CMSIS/CM3/DeviceSupport/ST/$Architecture/

vpath %.c additionals/250_stm_std_peripherals/src/ \\
          additionals/270_CPAL_CMSIS/CM3/DeviceSupport/ST/$Architecture/

vpath %.s additionals/270_CPAL_CMSIS/CM3/DeviceSupport/ST/$Architecture/startup/gcc_ride7

MAIN_OBJS += misc.o

END_OF_MAKEFILE
    createExtensionMakefileTail ${Name} $OldPWD/../extensions/ #}
    File="$OldPWD/../extensions/activate.${Name}.sh" #{ activate std_peripheral
    createActivateScriptHead "$File" "$0 1" "Architecture dependant standard peripheral library for $SubDir microcontrollers" 
    cat <<END_OF_ACTIVATE >>$File
#!/bin/bash

Module_ERROR=""

# remove existing links
rm 2>/dev/null \$Dir_ExtensionsActive/makefile.${Name}
rm 2>/dev/null \$Dir_ExtensionsActive/makefile.${INSTALLPATH}_*_StdPeriph_Driver


OldPWD=\`pwd\`
cd "\$Dir_Additionals"

rm 2>/dev/null 250_stm_std_peripherals
ln -s $Name2 250_stm_std_peripherals
cd "\$OldPWD"


# create links into extensions/
cd "\$Dir_Extensions"
rm 2>/dev/null *250_stm_std_peripherals_*

DriversFound=\`ls $Architecture/*\`
if [ "\$DriversFound" != "" ]; then
  ln -s $Architecture/* .
else
  Module_ERROR="Cannot find drivers in \`pwd\`/$Architecture/ !"
fi

cd "\$CurrentPath"
if [ "\$Module_ERROR" == "" ]; then
  # create links into extensions.active/
  createLink \$Dir_Extensions/makefile.$Name \$Dir_ExtensionsActive/makefile.$Name '' QUIET
else
  echo "\$0 - ERROR: \$Module_ERROR"
fi

END_OF_ACTIVATE
#}
    createActivateScriptTail $File ${Name}
    
    #{ create one activate-script per c-file
    TargetDir="../../../extensions/$Architecture"
    dir "$TargetDir"

    LibPrefix="250_stm_std_peripherals"
    for SourceFile in `find ${SubDir}/src/ -name "*\.c" -execdir echo -n "{} " \;` ; do #{
      SourceName=`perl -e "print substr('$SourceFile', 2, -2);"`
      Source=`perl -e "print substr('$SourceFile', 12, -2);"`
      if [ "$Source" != "" ]; then
        DestName="${LibPrefix}__${Source}"
        createExtensionMakefileHead ${DestName} $TargetDir/ #{
        cat <<END_OF_MAKEFILE >>$MakeFile # set by createExtensionMakefileHead() 
  
MAIN_OBJS += ${SourceName}.o

END_OF_MAKEFILE
        createExtensionMakefileTail ${DestName} $TargetDir/ #}
        File="$TargetDir/activate.${DestName}.sh" #{
        createActivateScriptHead $File $0 "adds c-source to list of compiled objects: ${SourceName}.c"
        cat <<END_OF_ACTIVATE >>$File #{

# add createL.inks for each link required for this extension
createLink ../extensions/makefile.$DestName \$Dir_ExtensionsActive/makefile.$DestName '' QUIET

END_OF_ACTIVATE
#}
        createActivateScriptTail $File $DestName
        #}
      fi
    done #}
    #} create one activate-script per c-file
  done #}
  
  #{ STM32_CPAL_Driver: create makefiles + activate scripts
  Device_IncludeDirs=`find STM32_CPAL_Driver/devices/ -mindepth 1 -type d -exec echo "                -I additionals/${INSTALLPATH}_{} \\\\" \;`
  Device_SourceDirs=`find STM32_CPAL_Driver/devices/ -mindepth 1 -type d -exec echo "          additionals/${INSTALLPATH}_{} \\\\" \;`
  Name="${INSTALLPATH}_STM32_CPAL_Driver"
  createExtensionMakefileHead ${Name} $OldPWD/../extensions/ #{ create makefile for STM32_CPAL_Driver
  cat <<END_OF_MAKEFILE >>$MakeFile

INCLUDE_DIRS += -I additionals/$INSTALLPATH/$Dir_CoreSupport
INCLUDE_DIRS += -I additionals/270_CPAL_STM32_CPAL_Driver/inc/ \\
$Device_IncludeDirs
                
                
vpath %.c additionals/$INSTALLPATH/$Dir_CoreSupport
vpath %.c additionals/270_CPAL_STM32_CPAL_Driver/src/ \\
$Device_SourceDirs


END_OF_MAKEFILE
  createExtensionMakefileTail ${Name} $OldPWD/../extensions/ #}
  File="$OldPWD/../extensions/activate.${Name}.sh"           #{ create activate script for STM32_CPAL_Driver
  createActivateScriptHead "$File" "$0 1" "Driver required to use architecture independant features of ST's Common Peripheral Access Library (CPAL)"
  cat <<END_OF_ACTIVATE >>$File #{
#!/bin/bash

#rm 2>/dev/null \$Dir_ExtensionsActive/makefile.${Name}_*

OldPWD=\`pwd\`
cd "\$Dir_Additionals"

cd "\$OldPWD"

# create links into extensions.active/
createLink \$Dir_Extensions/makefile.$Name \$Dir_ExtensionsActive/makefile.$Name '' QUIET

END_OF_ACTIVATE
#}
  createActivateScriptTail $File ${Name}
  #}
  
  #{ create one activate-script per c-file
  TargetDir="../../../extensions"
  #echo "TargetDir> ls `pwd`/$TargetDir"
  BaseName=$Name
  for SourceFile in `find STM32_CPAL_Driver/src/ -name "*\.c" -execdir echo -n "{} " \;` ; do #{
    SourceName=`perl -e "print substr('$SourceFile', 2, -2);"`
    if [ "$SourceName" != "" ]; then
      DestName="${BaseName}__${SourceName}"
      echo "new extension: $DestName"
      createExtensionMakefileHead ${DestName} $TargetDir/ #{
      cat <<END_OF_MAKEFILE >>$MakeFile # set by createExtensionMakefileHead() 

MAIN_OBJS += ${SourceName}.o

END_OF_MAKEFILE
      createExtensionMakefileTail ${DestName} $TargetDir/ #}
      File="$TargetDir/activate.${DestName}.sh" #{
      createActivateScriptHead $File $0 "adds c-source to list of compiled objects: ${SourceName}.c"
      cat <<END_OF_ACTIVATE >>$File #{

# activate CPAL
activate.${INSTALLPATH}_STM32_CPAL_Driver.sh QUIET \"\$0\"

# add createLinks for each link required for this extension
createLink ../extensions/makefile.$DestName \$Dir_ExtensionsActive/makefile.$DestName '' QUIET

END_OF_ACTIVATE
#}
      createActivateScriptTail $File $DestName
      #}
    fi
  done #}
  #}
  
  #}STM32_CPAL_Driver
  
  cd "$OldPWD"
  echo "Installed successfully: $INSTALLPATH"
#}
else                            #{ Install failed
  echo "failed to install $INSTALLPATH"
  exit 10
fi #}

cd ..
#} IMPLEMENT YOUR INSTALL SCRIPT ABOVE -----------------------------------------------

exit 0
